/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/*==============================================================================
  Leak Bucket Filter module
  
  Implements leaky bucket filter used at access point nodes for shaping the 
  traffic reveived from an application. 
  
 ============================================================================*/

#include "lbf.h"

// threshold of actual tokens; adjusting is done if at least one node's actual is
// more than GT_ADJ_ACTUAL_TH
static const float GT_ADJ_ACTUAL_TH = 25000;
//37500; 

// bucket for each node
static const int GT_ADJ_BUCKET = 25000; 

// rate is adjusted only if one underutilized node has this much excess capacity
static const float GT_ADJ_RATE_TH = 0.01;

static const int GT_ADJ_UNDER_CNT = 1;
static const int GT_ADJ_OVER_CNT = 1;



extern void lbf_item_init(struct lbf_item *it, struct lbf* lbf) 
{
	it->tokens = 50000; 
	it->bucket = 100000; 
#ifndef GT_DEV_ADJUST_TEST
	it->token_rate =  0.5 * lbf->max_token_rate;
#else
	it->token_rate =  0.5 * lbf->max_token_rate;
#endif

#ifdef GT_DEV_ADJUST
	it->bucket = GT_ADJ_BUCKET;
#endif
	it->actual = 0;
	it->under_cnt = 0;
	it->over_cnt = 0;
	it->max_buf_len = STA_MAX_TX_BUFFER;

	PDEBUG(DEBUG_LBF, "lbf_item_init: \n");
}


extern void lbf_init (struct lbf *lbf)
{
	lbf->firsttime = 1;
	lbf->use_tftime = 1; 
#ifndef GT_DEV_ADJUST_TEST
	lbf->max_token_rate = 1; //1 (all the tests used this)
#else
	lbf->max_token_rate = 0.805; //1 (all the tests used this)
#endif
	lbf->index = 0;
	spin_lock_init(&lbf->index_lock);
	spin_lock_init(&lbf->release_lock);

	memset(&lbf->lastadjust, 0, sizeof(lbf->lastadjust));
	memset(&lbf->lastupdate, 0, sizeof(lbf->lastupdate));
}	


// Update tokens of the station that transmitted or received SKB. 
// SKB has been sent (TO_SEND == 1) or received.  
extern void lbf_update_tokens(struct lbf *lbf, struct lbf_item *it, int tftime,
			      struct sta_info *sta, int sent)
{
	it->tokens = it->tokens - tftime;
#ifdef GT_DEV_ADJUST
	it->actual = it->actual + tftime;
#endif
	PDEBUG(DEBUG_LBF, "%ld update_tokens " MACSTR 
	       " sent %d, tftime %d, bucket %d, tokens %d\n", 
	       jiffies, MAC2STR(sta->addr), sent, tftime, it->bucket, it->tokens);
}

void lbf_update_index(struct lbf *lbf, int new_val)
{
	if (lbf && new_val >= 0) {
		spin_lock(&lbf->index_lock);
		lbf->index = new_val;
		spin_unlock(&lbf->index_lock);
		
	}
}

extern int lbf_ok_to_release(struct sta_info *sta, int size, struct ap_data *ap)
{
	int ret = 0;
#ifndef GT_DEV_ZERO_TK
	if (ap && sta && size <= sta->lbf_it.tokens) {
#else
	if (ap && sta && sta->lbf_it.tokens > 0) {
#endif
		struct list_head *ptr = NULL;
		struct sta_info *begin = NULL;
		int sta_cnt = 0;
		
		for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
			if (sta_cnt == ap->lbf->index) {
				begin = (struct sta_info*)ptr;
				if (begin == sta) {
					// continue
				}
				else {
#ifndef GT_DEV_DONT_WAIT_TURN
					if (skb_queue_len(&begin->tx_buf) > 0 && begin->lbf_it.tokens > 0) {
						// TODO: this only works if we only have two stations 
						// STA must wait for its turn
						PDEBUG(DEBUG_LBF, "%ld " MACSTR " must wait for its turn\n", 
						       jiffies, MAC2STR(sta->addr));
						goto out;
					}
#endif
				}
			}
			sta_cnt++;
		}

		if (sta_cnt > 0) {
			lbf_update_index(ap->lbf, (ap->lbf->index + 1) % sta_cnt);
		}
		else {
			// Shouldn't have been here
			PDEBUG(DEBUG_LBF_HIGH, "%ld lbf_ok_to_release SHOULDN't HAVE BEEN HERE\n", jiffies);
		}
		ret = 1;
	}
out:
	return ret;
}


// Release a packet from a suitable input queue. 
// NULL if no pkt is ready for release.
extern struct sk_buff* lbf_release(struct lbf *lbf, struct ap_data *ap)
{
	// round-robin should be good enough for coarse-grain fairness 
	// (won't be good when packet sizes differ a lot) 

	// TODO: The abstraction is a bit broken here. Lbf is not easily portable since
	// it uses STA
	int size = 0;
	int i = 0, sta_cnt = 0, buf_cnt = 0, qlen = 0;
	struct sta_info *sta;
	struct list_head *ptr, *begin = NULL;
	struct sk_buff *skb = NULL;

	//spin_lock(&lbf->release_lock);
	if (skb_queue_len(&ap->pri_tx_buf) > 0) {
		PDEBUG(DEBUG_LBF, "%ld lbf_release: dequeue from pri_txbuf\n", jiffies);
		skb = skb_dequeue(&ap->pri_tx_buf);
		//spin_unlock(&lbf->release_lock);
		return skb;
	}

	sta_cnt = 0;
	for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
		if (sta_cnt == lbf->index) {
			begin = ptr;
		}
		sta_cnt++;
	}

	i = 0; // how far you are from lbf->index
	for (ptr = begin; i < sta_cnt; i++) {
		sta = (struct sta_info*)ptr;
		if (sta)
			qlen = skb_queue_len(&sta->tx_buf);

		if (sta && qlen > 0) {
			buf_cnt += qlen;

			skb =  sta->tx_buf.next; // peep the first packet
			if (!skb) {
				PDEBUG(DEBUG_LBF_HIGH, "%ld PANIC! SKB IS NULL. qlen %d\n", jiffies, qlen);
				goto out;
			}
			size = hostap_tftime(skb->len, sta->tx_fixed_rate); // use 10 to assume the worst case tx_rate
#ifndef GT_DEV_ZERO_TK
			if (size <= sta->lbf_it.tokens) {
#else
			if (sta->lbf_it.tokens > 0) {
#endif
				skb = skb_dequeue(&sta->tx_buf);
				if (!skb) {
					PDEBUG(DEBUG_LBF_HIGH, "%ld " MACSTR " skb null. PANIC!\n", 
					       jiffies, MAC2STR(sta->addr));
				}
				PDEBUG(DEBUG_LBF, "%ld lbf_release: " MACSTR 
				       " DEQUE len %d, size %d, tokens %d, qlen %d,"
				       " buf_cnt %d, sta_cnt %d, index %d\n",
				       jiffies, MAC2STR(sta->addr), skb->len, size, 
				       sta->lbf_it.tokens, qlen, buf_cnt, sta_cnt,
				       (lbf->index + i) % sta_cnt);
				buf_cnt--;
				break;
			} 
			else {
				if (skb->len >= GT_DATA_SIZE) {
					PDEBUG(DEBUG_LBF, " %ld lbf_release: " MACSTR 
					       " NOTDE len %d, size %d, tokens %d, qlen %d, buf_cnt %d\n",
					       jiffies, MAC2STR(sta->addr), skb->len, 
					       size, sta->lbf_it.tokens, qlen, buf_cnt);
				}
				skb = NULL;
			}
		}

		ptr = ptr->next; // sta_list is a doubly linked list so it will go around when it reaches the end
		if (ptr == &ap->sta_list)
			ptr = ap->sta_list.next;
	}

 out:
	//spin_unlock(&lbf->release_lock);

	if (lbf && i < sta_cnt && i >= 0 && sta_cnt > 0) { //regard less of a pkt is found
		if (lbf->index < 0)
			lbf->index = 0;
		lbf_update_index(lbf, (lbf->index + i + 1) % sta_cnt);
	}

	if (buf_cnt <= 0) {
		ap->buffered = 0;
	}


	return skb;
}


// Update bucket of each LBF. 
extern void lbf_fill_tokens(struct lbf *lbf, struct ap_data *ap)
{
	struct timeval now;
	do_gettimeofday(&now); // overhead?

	if (lbf->firsttime) {
		lbf->firsttime = 0;
		lbf->lastupdate = now;
		return;
	}

	struct sta_info *sta;
	struct list_head *ptr;
	struct timeval diff = time_diff(&now, &lbf->lastupdate);
	int sz = diff.tv_usec; 
	if (diff.tv_sec > 0) {
		PDEBUG(DEBUG_LBF, "%ld lbf_fill_tokens: update interval TOO LONG sec %ld, usec %ld\n", 
		       jiffies, diff.tv_sec, diff.tv_usec);
		sz = 1000000;
	}

	for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
		sta = (struct sta_info*)ptr;
		sta->lbf_it.tokens += (int)(sz * sta->lbf_it.token_rate);
		if (sta->lbf_it.tokens > sta->lbf_it.bucket) {
			PDEBUG(DEBUG_LBF, "%ld " MACSTR " tokens %d more than bucket %d\n", 
			       jiffies, MAC2STR(sta->addr), sta->lbf_it.tokens - sta->lbf_it.bucket, sta->lbf_it.bucket);
			sta->lbf_it.tokens = sta->lbf_it.bucket;  // can't get more than the bucket
		}
	}

	PDEBUG(DEBUG_LBF, "%ld lbf_fill_tokens: adjusted interval %d usec\n",
	       jiffies, sz);
	lbf->lastupdate = now;
}


extern void lbf_adjust_rate(struct lbf *lbf, struct ap_data *ap)
{
	struct timeval now;
	do_gettimeofday(&now); // overhead?

	if (!lbf->lastadjust.tv_usec) {
		lbf->lastadjust = now;
		return;
	}
	
	static const int max_num_sta = 4;
	struct sta_info *over_stas[max_num_sta]; // over utilized stas
	int i;
	for (i = 0; i < max_num_sta; i++) {
		over_stas[i] = NULL;
	}
	int over_stas_index = 0;

	struct sta_info *sta, *sta_min = NULL;
	struct list_head *ptr;
	float actual_rate, d, d_min = 0;

 	for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
 		sta = (struct sta_info*)ptr;
 		// at least one STA must have used most of its bucket
 		if (sta->lbf_it.actual > GT_ADJ_ACTUAL_TH)
 			break;
 	}

 	if (ptr == &ap->sta_list)
 		return;

	struct timeval diff = time_diff(&now, &lbf->lastadjust);
	int intv = diff.tv_usec; 
	if (diff.tv_sec > 0) {
		PDEBUG(DEBUG_LBF_HIGH, "%ld lbf_adjust_tokens: update interval TOO LONG sec %ld, usec %ld\n", 
		       jiffies, diff.tv_sec, diff.tv_usec);
		intv = 1000000;
	}

	//return; //@GT-1 TESTING 


	for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
		sta = (struct sta_info*)ptr;
		actual_rate = (1.0 * sta->lbf_it.actual) / intv;
		d = sta->lbf_it.token_rate - actual_rate; // excess capacity
		if (d > GT_ADJ_RATE_TH) {
			//&& sta->lbf_it.actual >= GT_ADJ_ACTUAL_TH) {
			sta->lbf_it.over_cnt = 0; 
			if (sta->lbf_it.under_cnt++ < GT_ADJ_UNDER_CNT) {
				continue;
			}

			if (sta->lbf_it.actual > 0) {
				PDEBUG(DEBUG_LBF, "%ld " MACSTR " SMALLER actual_rate %d, rate %d, token %d, undercnt %d\n",
				       jiffies, MAC2STR(sta->addr), 
				       FL2INT(actual_rate), 
				       FL2INT(sta->lbf_it.token_rate), 
				       sta->lbf_it.tokens, 
				       sta->lbf_it.under_cnt);
			}

			if (d_min <= 0 || d < d_min) {
				d_min  = d;
				sta_min = sta;
			}
		}
		else {
			sta->lbf_it.under_cnt = 0; 

			if (sta->lbf_it.over_cnt++ >= GT_ADJ_OVER_CNT) {
				if (over_stas_index < max_num_sta) {
					over_stas[over_stas_index++] = sta;
				}
				else {
					PDEBUG(DEBUG_LBF, "%ld max_num_sta is too small! PANIC!\n", jiffies);
				}

				//if (d <= (-1 * GT_ADJ_RATE_TH)) {
				PDEBUG(DEBUG_LBF, "%ld " MACSTR " BIGGER  actual_rate %d, rate %d, token %d, overcnt %d\n",
				       jiffies,  MAC2STR(sta->addr), 
				       FL2INT(actual_rate), 
				       FL2INT(sta->lbf_it.token_rate),
				       sta->lbf_it.tokens,
				       sta->lbf_it.over_cnt);
			}
		}
	}

	if (sta_min && over_stas_index) {
		d_min = (d_min / 2.0);
		float incr = d_min / over_stas_index;
		
		for (i = 0; i < over_stas_index; i++) {
			sta = over_stas[i];

			PDEBUG(DEBUG_LBF_HIGH, "%ld " MACSTR " rate %4d INCR by %4d, actrate %4d," 
			       " tok %6d, acttok %6d, under %d, over %d\n", 
			       jiffies, MAC2STR(sta->addr), 
			       FL2INT(sta->lbf_it.token_rate), 
			       FL2INT(incr), 
			       FL2INT(1.0 * sta->lbf_it.actual / intv), 
			       sta->lbf_it.tokens, 
			       sta->lbf_it.actual, 
			       sta->lbf_it.under_cnt,
			       sta->lbf_it.over_cnt);

			sta->lbf_it.token_rate += incr;
		}
		
		PDEBUG(DEBUG_LBF_HIGH, "%ld " MACSTR " rate %4d DECR by %4d, actrate %4d," 
		       " tok %6d, acttok %6d, under %d, over %d\n", 
		       jiffies, MAC2STR(sta_min->addr), 
		       FL2INT(sta_min->lbf_it.token_rate), 
		       FL2INT(d_min),
		       FL2INT(1.0 * sta_min->lbf_it.actual / intv), 
		       sta_min->lbf_it.tokens, 
		       sta_min->lbf_it.tokens, 
		       sta_min->lbf_it.under_cnt,
		       sta_min->lbf_it.over_cnt);

		sta_min->lbf_it.token_rate -= incr;

		for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
			sta = (struct sta_info*)ptr;
			sta->lbf_it.actual = 0; 
			// reset under_cnt and over_cnt is rate is now being adjusted
			sta->lbf_it.under_cnt = 0; 
			sta->lbf_it.over_cnt = 0; 

		}
	}

	// reset this for all nodes even if no adjustment occurs
	for (ptr = ap->sta_list.next; ptr != &ap->sta_list; ptr = ptr->next) {
		sta = (struct sta_info*)ptr;
		sta->lbf_it.actual = 0; 
	}

	PDEBUG(DEBUG_LBF, "%ld lbf_adjust_tokens: adjusted interval %d usec\n",
	       jiffies, intv);
	lbf->lastadjust = now;
}


